from typing import Any, Optional, cast
from typing_extensions import Protocol, runtime_checkable

from pydantic import Field
from nonebot.log import logger
from nonebot.typing import overrides

from nonebot.adapters import Event as BaseEvent

from .model import *
from .message import Message


@runtime_checkable
class EventWithChat(Protocol):
    chat: Chat


class Event(BaseEvent):
    @classmethod
    def __parse_event(cls, obj: dict) -> "Event":
        logger.debug('"RAW":' + str(obj))
        post_type: str = list(obj.keys())[1]
        event_map = {
            "message": MessageEvent,
            "edited_message": EditedMessageEvent,
            "channel_post": MessageEvent,
            "edited_channel_post": EditedMessageEvent,
            "inline_query": InlineQueryEvent,
            "chosen_inline_result": ChosenInlineResult,
            "callback_query": CallbackQueryEvent,
            "shipping_query": ShippingQueryEvent,
            "pre_checkout_query": PreCheckoutQueryEvent,
            "poll": PollEvent,
            "poll_answer": PollAnswerEvent,
            "chat_member": ChatMemberUpdatedEvent,
            "my_chat_member": ChatMemberUpdatedEvent,
        }
        return event_map[post_type].parse_event(obj[post_type])

    @classmethod
    def _parse_event(cls, obj: dict, failed: set = set()) -> "Event":
        for subclass in cls.__subclasses__():
            if subclass not in failed:
                try:
                    return subclass.parse_event(obj)
                except:
                    pass
        return cls.parse_obj(obj)

    @classmethod
    def parse_event(cls, obj: dict) -> "Event":
        if hasattr(cls, f"_{cls.__name__}__parse_event"):
            return getattr(cls, f"_{cls.__name__}__parse_event")(obj)
        else:
            return cls._parse_event(obj)

    @overrides(BaseEvent)
    def get_type(self) -> str:
        return ""

    @overrides(BaseEvent)
    def get_event_name(self) -> str:
        return ""

    @overrides(BaseEvent)
    def get_event_description(self) -> str:
        return str(self.dict(by_alias=True, exclude_none=True))

    @overrides(BaseEvent)
    def get_user_id(self) -> str:
        raise ValueError("Event has no user!")

    @overrides(BaseEvent)
    def get_session_id(self) -> str:
        raise ValueError("Event has no session!")

    @overrides(BaseEvent)
    def get_message(self) -> Message:
        raise ValueError("Event has no message!")

    @overrides(BaseEvent)
    def get_plaintext(self) -> str:
        raise ValueError("Event has no message!")

    @overrides(BaseEvent)
    def is_tome(self) -> bool:
        return False


class MessageEvent(Event):
    message_id: int
    date: int
    chat: Chat
    forward_from: Optional[User]
    forward_from_chat: Optional[Chat]
    forward_from_message_id: Optional[int]
    forward_signature: Optional[str]
    forward_sender_name: Optional[str]
    via_bot: Optional[User]
    media_group_id: Optional[str]
    author_signature: Optional[str]
    reply_to_message: Optional["MessageEvent"] = None
    message: Message = Message()

    @classmethod
    def __parse_event(cls, obj: dict) -> "Event":
        if not Message(obj):
            return NoticeEvent.parse_event(obj)
        else:
            message_type = obj["chat"]["type"]
            event_map = {
                "private": PrivateMessageEvent,
                "group": GroupMessageEvent,
                "supergroup": GroupMessageEvent,
                "channel": ChannelPostEvent,
            }
            return event_map[message_type].parse_event(obj)

    def __init__(self, **data: Any) -> None:
        reply_to_message = data.pop("reply_to_message", None)
        super().__init__(**data)
        self.message = Message(data)
        if reply_to_message:
            self.reply_to_message = cast(
                MessageEvent, self.parse_event(reply_to_message)
            )

    @overrides(Event)
    def get_type(self) -> str:
        return "message"

    @overrides(Event)
    def get_event_name(self) -> str:
        return "message"

    @overrides(Event)
    def get_message(self) -> Message:
        return self.message

    @overrides(Event)
    def get_plaintext(self) -> str:
        return self.message.extract_plain_text()

    # TODO
    @overrides(Event)
    def is_tome(self) -> bool:
        return super().is_tome()


class PrivateMessageEvent(MessageEvent):
    from_: User = Field(alias="from")

    @overrides(MessageEvent)
    def get_event_name(self) -> str:
        return "message.private"

    @overrides(MessageEvent)
    def get_user_id(self) -> str:
        return str(self.from_.id)

    @overrides(MessageEvent)
    def get_session_id(self) -> str:
        return f"private_{self.chat.id}"

    @overrides(MessageEvent)
    def is_tome(self) -> bool:
        return True


class GroupMessageEvent(MessageEvent):
    from_: User = Field(alias="from")
    sender_chat: Optional[Chat]

    @overrides(MessageEvent)
    def get_event_name(self) -> str:
        return "message.group"

    @overrides(MessageEvent)
    def get_user_id(self) -> str:
        return str(self.from_.id)

    @overrides(MessageEvent)
    def get_session_id(self) -> str:
        return f"group_{self.chat.id}_{self.from_.id}"


class ChannelPostEvent(MessageEvent):
    sender_chat: Optional[Chat]

    @overrides(MessageEvent)
    def get_event_name(self) -> str:
        return "message.channel_post"

    @overrides(MessageEvent)
    def get_session_id(self) -> str:
        return f"channel_{self.chat.id}"


class EditedMessageEvent(Event):
    message_id: int
    date: int
    chat: Chat
    via_bot: Optional[User]
    edit_date: int
    media_group_id: Optional[str]
    author_signature: Optional[str]
    reply_to_message: Optional["MessageEvent"]
    message: Message = Message()

    @classmethod
    def __parse_event(cls, obj: dict):
        message_type = obj["chat"]["type"]
        event_map = {
            "private": PrivateEditedMessageEvent,
            "group": GroupEditedMessageEvent,
            "supergroup": GroupEditedMessageEvent,
            "channel": EditedChannelPostEvent,
        }
        return event_map[message_type].parse_event(obj)

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.message = Message(data)

    @overrides(Event)
    def get_type(self) -> str:
        return "edit_message"

    @overrides(Event)
    def get_event_name(self) -> str:
        return "edit_message"

    @overrides(Event)
    def get_message(self) -> Message:
        return self.message

    @overrides(Event)
    def get_plaintext(self) -> str:
        return self.message.extract_plain_text()

    # TODO
    @overrides(Event)
    def is_tome(self) -> bool:
        return super().is_tome()


class PrivateEditedMessageEvent(EditedMessageEvent):
    from_: User = Field(alias="from")
    sender_chat: Optional[Chat]

    @overrides(EditedMessageEvent)
    def get_event_name(self) -> str:
        return "edited_message.private"

    @overrides(EditedMessageEvent)
    def get_user_id(self) -> str:
        return str(self.from_.id)

    @overrides(EditedMessageEvent)
    def get_session_id(self) -> str:
        return f"private_{self.chat.id}"

    @overrides(EditedMessageEvent)
    def is_tome(self) -> bool:
        return True


class GroupEditedMessageEvent(EditedMessageEvent):
    from_: User = Field(default=None, alias="from")
    sender_chat: Optional[Chat]

    @overrides(EditedMessageEvent)
    def get_event_name(self) -> str:
        return f"edited_message.group"

    @overrides(EditedMessageEvent)
    def get_user_id(self) -> str:
        return str(self.from_.id)

    @overrides(EditedMessageEvent)
    def get_session_id(self) -> str:
        return f"group_{self.chat.id}_{self.from_.id}"


class EditedChannelPostEvent(EditedMessageEvent):
    sender_chat: Optional[Chat]

    @overrides(EditedMessageEvent)
    def get_event_name(self) -> str:
        return "edited_message.channel_post"

    @overrides(EditedMessageEvent)
    def get_session_id(self) -> str:
        return f"channel_{self.chat.id}"


class NoticeEvent(Event):
    @classmethod
    def __parse_event(cls, obj: dict) -> "Event":
        if "pinned_message" in obj:
            return PinnedMessageEvent.parse_event(obj)
        elif "new_chat_member" in obj:
            return NewChatMemberEvent.parse_event(obj)
        elif "left_chat_member" in obj:
            return LeftChatMemberEvent.parse_event(obj)
        else:
            return cls._parse_event(obj)

    @overrides(Event)
    def get_type(self) -> str:
        return "notice"

    @overrides(Event)
    def get_event_name(self) -> str:
        return "notice"


class PinnedMessageEvent(NoticeEvent):
    message_id: int
    from_: User = Field(alias="from")
    sender_chat: Optional[Chat]
    chat: Chat
    date: int
    pinned_message: MessageEvent

    @overrides(NoticeEvent)
    def get_event_name(self) -> str:
        return "notice.pinned_message"

    @overrides(NoticeEvent)
    def get_message(self) -> Message:
        return self.pinned_message.get_message()

    @overrides(NoticeEvent)
    def get_plaintext(self) -> str:
        return self.pinned_message.get_plaintext()


class NewChatMemberEvent(NoticeEvent):
    message_id: int
    from_: Optional[User] = Field(default=None, alias="from")
    chat: Chat
    date: int
    new_chat_participant: Optional[User]
    new_chat_member: User
    new_chat_members: Optional[List[User]]

    @overrides(NoticeEvent)
    def get_event_name(self) -> str:
        return "notice.chat_member.new"


class LeftChatMemberEvent(NoticeEvent):
    message_id: int
    from_: Optional[User] = Field(default=None, alias="from")
    chat: Chat
    date: int
    left_chat_participant: Optional[User]
    left_chat_member: User

    @overrides(Event)
    def get_event_name(self) -> str:
        return "notice.chat_member.left"


class ChatMemberUpdatedEvent(NoticeEvent):
    chat: Chat
    from_: User = Field(alias="from")
    date: int
    old_chat_member: ChatMember
    new_chat_member: ChatMember
    invite_link: Optional[ChatInviteLink]

    @overrides(Event)
    def get_event_name(self) -> str:
        return "notice.chat_member.updated"


# TODO
class InlineQueryEvent(Event):
    id: str
    from_: User = Field(alias="from")
    query: str
    offset: str
    chat_type: Optional[str]
    Location: Optional[Location]


# TODO
class ChosenInlineResultEvent(Event):
    result_id: str
    from_: User = Field(alias="from")
    Location: Optional[Location]
    inline_message_id: Optional[str]
    query: str


# TODO
class CallbackQueryEvent(Event):
    id: str
    from_: Optional[User] = Field(default=None, alias="from")
    message: Optional[Message_]
    inline_message_id: Optional[str]
    chat_instance: Optional[str]
    data: Optional[str]
    game_short_name: Optional[str]


# DELAY
class ShippingQueryEvent(Event):
    id: str
    from_: User = Field(alias="from")
    invoice_payload: str
    shipping_address: ShippingAddress


# DELAY
class PreCheckoutQueryEvent(Event):
    id: str
    from_: User = Field(alias="from")
    currency: str
    total_amount: int
    invoice_payload: str
    shipping_option_id: Optional[str]
    order_info: Optional[OrderInfo]


# TODO
class PollEvent(Event):
    id: str
    question: str
    options: List[PollOption]
    total_voter_count: int
    is_closed: bool
    is_anonymous: bool
    type: str
    allows_multiple_answers: bool
    correct_option_id: Optional[int]
    explanation: Optional[str]
    explanation_entities: Optional[List[MessageEntity]]
    open_period: Optional[int]
    close_date: Optional[int]


# TODO
class PollAnswerEvent(Event):
    poll_id: str
    user: User
    option_ids: List[int]



from typing import Any, List, Union, Optional

from nonebot.adapters import Bot as BaseBot
from nonebot.adapters import Adapter as Adapter

from .model import *
from .event import Event as Event
from .message import Message as Message
from .config import BotConfig as BotConfig
from .event import EventWithChat as EventWithChat
from .message import MessageSegment as MessageSegment

class Bot(BaseBot):
    adapter: Adapter
    self_id: str
    bot_config: BotConfig
    def __init__(self, adapter: Adapter, config: BotConfig) -> None: ...
    async def send(
        self, event: Event, message: Union[str, Message, MessageSegment], **kwargs
    ) -> Any: ...
    async def get_updates(
        self,
        offset: Optional[int],
        limit: Optional[int],
        timeout: Optional[int],
        allowed_updates: Optional[List[str]],
    ): ...
    async def set_webhook(
        self,
        url: str,
        certificate: Optional[InputFile],
        ip_address: Optional[str],
        max_connections: Optional[int],
        allowed_updates: Optional[List[str]],
        drop_pending_updates: Optional[bool],
    ): ...
    async def delete_webhook(self, drop_pending_updates: Optional[bool]): ...
    async def get_webhook_info(self): ...
    async def get_me(self):
        """
        :说明:
          用于测试机器人 Token 的 API
        :返回:
          * ``User``: 机器人本身的 User
        """
        ...
    async def log_out(self): ...
    async def close(self): ...
    async def send_message(
        self,
        chat_id: Union[int, str],
        text: str,
        parse_mode: Optional[str],
        entities: Optional[List[MessageEntity]],
        disable_web_page_preview: Optional[bool],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def forward_message(
        self,
        chat_id: Union[int, str],
        from_chat_id: Union[int, str],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        message_id: int,
    ): ...
    async def copy_message(
        self,
        chat_id: Union[int, str],
        from_chat_id: Union[int, str],
        message_id: int,
        caption: Optional[str],
        parse_mode: Optional[str],
        caption_entities: Optional[List[MessageEntity]],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_photo(
        self,
        chat_id: Union[int, str],
        photo: Union[InputFile, str],
        caption: Optional[str],
        parse_mode: Optional[str],
        caption_entities: Optional[List[MessageEntity]],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_audio(
        self,
        chat_id: Union[int, str],
        audio: Union[InputFile, str],
        caption: Optional[str],
        parse_mode: Optional[str],
        caption_entities: Optional[List[MessageEntity]],
        duration: Optional[int],
        performer: Optional[str],
        title: Optional[str],
        thumb: Optional[Union[InputFile, str]],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_document(
        self,
        chat_id: Union[int, str],
        document: Union[InputFile, str],
        thumb: Optional[Union[InputFile, str]],
        caption: Optional[str],
        parse_mode: Optional[str],
        caption_entities: Optional[List[MessageEntity]],
        disable_content_type_detection: Optional[bool],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_video(
        self,
        chat_id: Union[int, str],
        video: Union[InputFile, str],
        duration: Optional[int],
        width: Optional[int],
        height: Optional[int],
        thumb: Optional[Union[InputFile, str]],
        caption: Optional[str],
        parse_mode: Optional[str],
        caption_entities: Optional[List[MessageEntity]],
        support_streaming: Optional[bool],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_animation(
        self,
        chat_id: Union[int, str],
        animation: Union[InputFile, str],
        duration: Optional[int],
        width: Optional[int],
        height: Optional[int],
        thumb: Optional[Union[InputFile, str]],
        caption: Optional[str],
        parse_mode: Optional[str],
        caption_entities: Optional[List[MessageEntity]],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_voice(
        self,
        chat_id: Union[int, str],
        voice: Union[InputFile, str],
        caption: Optional[str],
        parse_mode: Optional[str],
        caption_entities: Optional[List[MessageEntity]],
        duration: Optional[int],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_video_note(
        self,
        chat_id: Union[int, str],
        video_note: Union[InputFile, str],
        duration: Optional[int],
        length: Optional[int],
        thumb: Optional[Union[InputFile, str]],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_media_group(
        self,
        chat_id: Union[int, str],
        media: List[
            Union[InputMediaAudio, InputMediaDocument, InputMediaPhoto, InputMediaVideo]
        ],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
    ): ...
    async def send_location(
        self,
        chat_id: Union[int, str],
        latitude: float,
        longitude: float,
        horizontal_accuracy: Optional[float],
        live_period: Optional[int],
        heading: Optional[int],
        proximity_alert_radius: Optional[int],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def edit_message_live_location(
        self,
        chat_id: Optional[Union[int, str]],
        message_id: Optional[int],
        inline_message_id: Optional[int],
        latitude: float,
        longitude: float,
        horizontal_accuracy: Optional[float],
        heading: Optional[int],
        proximity_alert_radius: Optional[int],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def stop_message_live_location(
        self,
        chat_id: Optional[Union[int, str]],
        message_id: Optional[int],
        inline_message_id: Optional[int],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_venue(
        self,
        chat_id: Union[int, str],
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        foursquare_id: Optional[str],
        foursquare_type: Optional[str],
        google_place_id: Optional[str],
        google_place_type: Optional[str],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_contact(
        self,
        chat_id: Union[int, str],
        phone_number: str,
        first_name: str,
        last_name: Optional[str],
        vcard: Optional[str],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_poll(
        self,
        chat_id: Union[int, str],
        question: str,
        options: List[str],
        is_anonymous: Optional[bool],
        type: Optional[str],
        allows_multiple_answers: Optional[bool],
        correct_option_id: Optional[int],
        explanation: Optional[str],
        explanation_parse_mode: Optional[str],
        explanation_entities: Optional[List[MessageEntity]],
        open_period: Optional[int],
        close_date: Optional[int],
        is_closed: Optional[bool],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_dice(
        self,
        chat_id: Union[int, str],
        emoji: Optional[str],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def send_chat_action(self, chat_id: Union[int, str], action: str): ...
    async def get_user_profile_photos(
        self, user_id: int, offset: Optional[int], limit: Optional[int]
    ): ...
    async def get_file(self, file_id: str): ...
    async def ban_chat_member(
        self,
        chat_id: Union[int, str],
        user_id: int,
        until_date: Optional[int],
        revoke_messages: Optional[bool],
    ): ...
    async def unban_chat_member(
        self, chat_id: Union[int, str], user_id: int, only_if_banned: Optional[bool]
    ): ...
    async def promote_chat_member(
        self,
        chat_id: Union[int, str],
        user_id: int,
        is_anonymous: Optional[bool],
        can_manage_chat: Optional[bool],
        can_post_messages: Optional[bool],
        can_edit_messages: Optional[bool],
        can_delete_messages: Optional[bool],
        can_manage_voice_chats: Optional[bool],
        can_restrict_members: Optional[bool],
        can_promote_members: Optional[bool],
        can_change_info: Optional[bool],
        can_invite_users: Optional[bool],
        can_pin_messages: Optional[bool],
    ): ...
    async def set_chat_administrator_custom_title(
        self, chat_id: Union[int, str], user_id: int, custom_title: str
    ): ...
    async def ban_chat_sender_chat(
        self, chat_id: Union[int, str], sender_chat_id: int
    ): ...
    async def unban_chat_sender_chat(
        self, chat_id: Union[int, str], sender_chat_id: int
    ): ...
    async def set_chat_permissions(
        self, chat_id: Union[int, str], permissions: ChatPermissions
    ): ...
    async def export_chat_invite_link(self, chat_id: Union[int, str]): ...
    async def create_chat_invite_link(
        self,
        chat_id: Union[int, str],
        name: Optional[str],
        expire_date: Optional[int],
        member_limit: Optional[int],
        creates_join_request: Optional[bool],
    ): ...
    async def edit_chat_invite_link(
        self,
        chat_id: Union[int, str],
        invite_link: str,
        name: Optional[str],
        expire_date: Optional[int],
        member_limit: Optional[int],
        creates_join_request: Optional[bool],
    ): ...
    async def revoke_chat_invite_link(
        self, chat_id: Union[int, str], invite_link: str
    ): ...
    async def approve_chat_join_request(
        self, chat_id: Union[int, str], user_id: int
    ): ...
    async def decline_chat_join_request(
        self, chat_id: Union[int, str], user_id: int
    ): ...
    async def set_chat_photo(self, chat_id: Union[int, str], photo: InputFile): ...
    async def delete_chat_photo(self, chat_id: Union[int, str]): ...
    async def set_chat_title(self, chat_id: Union[int, str], title: str): ...
    async def set_chat_description(
        self, chat_id: Union[int, str], description: Optional[str]
    ): ...
    async def pin_chat_message(
        self,
        chat_id: Union[int, str],
        message_id: int,
        disable_notification: Optional[bool],
    ): ...
    async def unpin_chat_message(
        self, chat_id: Union[int, str], message_id: Optional[int]
    ): ...
    async def unpin_all_chat_messages(self, chat_id: Union[int, str]): ...
    async def leave_chat(self, chat_id: Union[int, str]): ...
    async def get_chat(self, chat_id: Union[int, str]): ...
    async def get_chat_administrators(self, chat_id: Union[int, str]): ...
    async def get_chat_member_count(self, chat_id: Union[int, str]): ...
    async def get_chat_member(self, chat_id: Union[int, str], user_id: int): ...
    async def set_chat_sticker_set(
        self, chat_id: Union[int, str], sticker_set_name: str
    ): ...
    async def delete_chat_sticker_set(self, chat_id: Union[int, str]): ...
    async def answer_callback_query(
        self,
        callback_query_id: str,
        text: Optional[str],
        show_alert: Optional[bool],
        url: Optional[str],
        cache_time: Optional[int],
    ): ...
    async def set_my_commands(
        self,
        commands: List[BotCommand],
        scope: Optional[BotCommandScope],
        language_code: Optional[str],
    ): ...
    async def delete_my_commands(
        self, scope: Optional[BotCommandScope], language_code: Optional[str]
    ): ...
    async def get_my_commands(
        self, scope: Optional[BotCommandScope], language_code: Optional[str]
    ): ...
    async def edit_message_text(
        self,
        chat_id: Optional[Union[int, str]],
        message_id: Optional[int],
        inline_message_id: Optional[str],
        text: str,
        parse_mode: Optional[str],
        entities: Optional[List[MessageEntity]],
        disable_web_page_preview: Optional[bool],
        reply_markup: Optional[InlineKeyboardMarkup],
    ): ...
    async def edit_message_caption(
        self,
        chat_id: Optional[Union[int, str]],
        message_id: Optional[int],
        inline_message_id: Optional[str],
        caption: str,
        parse_mode: Optional[str],
        caption_entities: Optional[List[MessageEntity]],
        reply_markup: Optional[InlineKeyboardMarkup],
    ): ...
    async def edit_message_media(
        self,
        chat_id: Optional[Union[int, str]],
        message_id: Optional[int],
        inline_message_id: Optional[str],
        media: InputMedia,
        reply_markup: Optional[InlineKeyboardMarkup],
    ): ...
    async def edit_message_reply_markup(
        self,
        chat_id: Optional[Union[int, str]],
        message_id: Optional[int],
        inline_message_id: Optional[str],
        reply_markup: Optional[InlineKeyboardMarkup],
    ): ...
    async def stop_poll(
        self,
        chat_id: Optional[Union[int, str]],
        message_id: Optional[int],
        reply_markup: Optional[InlineKeyboardMarkup],
    ): ...
    async def delete_message(
        self, chat_id: Optional[Union[int, str]], message_id: Optional[int]
    ): ...
    async def send_sticker(
        self,
        chat_id: Union[int, str],
        sticker: Union[InputFile, str],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[
            Union[
                InlineKeyboardMarkup,
                ReplyKeyboardMarkup,
                ReplyKeyboardRemove,
                ForceReply,
            ]
        ],
    ): ...
    async def get_sticker_set(self, name: str): ...
    async def upload_sticker_file(self, user_id: int, png_sticker: InputFile): ...
    async def create_new_sticker_set(
        self,
        user_id: int,
        name: str,
        title: str,
        png_sticker: Optional[Union[InputFile, str]],
        tgs_sticker: Optional[InputFile],
        emojis: str,
        contains_masks: Optional[bool],
        mask_potion: Optional[MaskPosition],
    ): ...
    async def add_sticker_to_set(
        self,
        user_id: int,
        name: str,
        png_sticker: Optional[Union[InputFile, str]],
        tgs_sticker: Optional[InputFile],
        emojis: str,
        mask_potion: Optional[MaskPosition],
    ): ...
    async def set_sticker_position_in_set(self, sticker: str, position: int): ...
    async def delete_sticker_from_set(self, sticker: str): ...
    async def set_sticker_set_thumb(
        self, name: str, user_id: int, thumb: Optional[Union[InputFile, str]]
    ): ...
    async def answer_inline_query(
        self,
        inline_query_id: str,
        results: List[InlineQueryResult],
        cache_time: Optional[int],
        is_personal: Optional[bool],
        next_offset: Optional[str],
        switch_pm_text: Optional[str],
        switch_pm_parameter: Optional[str],
    ): ...
    async def send_invoice(
        self,
        chat_id: Union[int, str],
        title: str,
        description: str,
        payload: str,
        provider_token: str,
        prices: List[LabeledPrice],
        max_tip_amount: Optional[int],
        suggested_tip_amounts: Optional[int],
        start_parameter: Optional[str],
        provider_date: Optional[str],
        photo_url: Optional[str],
        photo_size: Optional[int],
        photo_width: Optional[int],
        photo_height: Optional[int],
        need_name: Optional[bool],
        need_phone_number: Optional[bool],
        need_email: Optional[bool],
        need_shipping_address: Optional[bool],
        send_phone_number_to_provider: Optional[bool],
        send_email_to_provider: Optional[bool],
        is_flexible: Optional[bool],
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[InlineKeyboardMarkup],
    ): ...
    async def answer_shipping_query(
        self,
        shipping_query_id: str,
        ok: bool,
        shipping_options: Optional[List[ShippingOption]],
        error_message: Optional[str],
    ): ...
    async def answer_pre_checkout_query(
        self, pre_checkout_query_id: str, ol: bool, error_message: Optional[str]
    ): ...
    async def set_passport_data_errors(
        self, user_id: int, errors: List[PassportElementError]
    ): ...
    async def send_game(
        self,
        chat_id: int,
        game_short_name: str,
        disable_notification: Optional[bool],
        protect_content: Optional[bool],
        reply_to_message_id: Optional[int],
        allow_sending_without_reply: Optional[bool],
        reply_markup: Optional[InlineKeyboardMarkup],
    ): ...
    async def set_game_score(
        self,
        user_id: int,
        score: int,
        force: Optional[bool],
        disable_edit_message: Optional[bool],
        chat_id: Optional[int],
        message_id: Optional[int],
        inline_message_id: Optional[str],
    ): ...
    async def get_game_high_score(
        self,
        user_id: int,
        chat_id: Optional[int],
        message_id: Optional[int],
        inline_message_id: Optional[str],
    ): ...



from typing import Any, Union

from nonebot.typing import overrides

from nonebot.adapters import Adapter
from nonebot.adapters import Bot as BaseBot

from .config import BotConfig
from .event import Event, EventWithChat
from .message import Message, MessageSegment


class Bot(BaseBot):
    """
    Telegram Bot 适配。继承属性参考 `BaseBot <./#class-basebot>`_ 。
    """

    def __init__(self, adapter: "Adapter", config: BotConfig):
        self.adapter = adapter
        self.self_id = config.token.split(":")[0]
        self.bot_config = config

    @overrides(BaseBot)
    async def send(
        self,
        event: Event,
        message: Union[str, Message, MessageSegment],
        **kwargs,
    ) -> Any:
        """
        由于 Telegram 对于不同类型的消息有不同的 API，如果需要批量发送不同类型的消息请尽量使用此方法，Nonebot 将会自动帮你转换成多条消息。
        """
        if isinstance(event, EventWithChat):
            if isinstance(message, str):
                return await self.send_message(chat_id=event.chat.id, text=message)
            elif isinstance(message, MessageSegment):
                if message.type == "text":
                    return await self.send_message(
                        chat_id=event.chat.id, text=message.data.get("text")
                    )
                elif message.type in [
                    "animation",
                    "audio",
                    "document",
                    "photo",
                    "video",
                    "voice",
                ]:
                    return await self.call_api(
                        f"send_{message.type}",
                        chat_id=event.chat.id,
                        **{
                            message.type: message.data.get("file"),
                            "caption": message.data.get("caption"),
                        },
                    )
            else:
                for seg in message:
                    if seg.type == "text":
                        return await self.send_message(
                            chat_id=event.chat.id, text=seg.data.get("text")
                        )



from typing import Any, Type, Union, Mapping, Iterable, Optional, cast

from nonebot.typing import overrides

from nonebot.adapters import Message as BaseMessage
from nonebot.adapters import MessageSegment as BaseMessageSegment


class MessageSegment(BaseMessageSegment):
    """
    telegram 协议 MessageSegment 适配。具体方法参考协议消息段类型或源码。
    """

    @classmethod
    @overrides(BaseMessageSegment)
    def get_message_class(cls) -> Type["Message"]:
        return Message

    @overrides(BaseMessageSegment)
    def __str__(self) -> str:
        if self.type == "text":
            return self.data["text"]
        return ""

    @overrides(BaseMessageSegment)
    def is_text(self) -> bool:
        return self.type == "text"

    @staticmethod
    def text(text: str) -> "MessageSegment":
        return MessageSegment("text", {"text": text})

    @staticmethod
    def photo(file: str, caption: Optional[str] = None) -> "MessageSegment":
        return MessageSegment("photo", {"file": file, "caption": caption})

    @staticmethod
    def voice(file: str, caption: Optional[str] = None) -> "MessageSegment":
        return MessageSegment("voice", {"file": file, "caption": caption})

    @staticmethod
    def animation(
        file: str, thumb: Optional[str] = None, caption: Optional[str] = None
    ) -> "MessageSegment":
        return MessageSegment(
            "animation", {"file": file, "thumb": thumb, "caption": caption}
        )

    @staticmethod
    def audio(
        file: str, thumb: Optional[str] = None, caption: Optional[str] = None
    ) -> "MessageSegment":
        return MessageSegment(
            "audio", {"file": file, "thumb": thumb, "caption": caption}
        )

    @staticmethod
    def document(
        file: str, thumb: Optional[str] = None, caption: Optional[str] = None
    ) -> "MessageSegment":
        return MessageSegment(
            "document", {"file": file, "thumb": thumb, "caption": caption}
        )

    @staticmethod
    def video(
        file: str, thumb: Optional[str] = None, caption: Optional[str] = None
    ) -> "MessageSegment":
        return MessageSegment(
            "video", {"file": file, "thumb": thumb, "caption": caption}
        )


class Message(BaseMessage[MessageSegment]):
    @classmethod
    @overrides(BaseMessage)
    def get_segment_class(cls) -> Type["MessageSegment"]:
        return MessageSegment

    @staticmethod
    @overrides(BaseMessage)
    def _construct(
        msg: Union[str, Mapping, Iterable[Mapping]]
    ) -> Iterable[MessageSegment]:
        # TODO
        if isinstance(msg, Mapping):
            msg = cast(Mapping[str, Any], msg)
            for key in msg:
                if key == "text":
                    yield MessageSegment(
                        key, {key: msg[key], "entities": msg.get("entities")}
                    )
                elif key == "photo":
                    yield MessageSegment(
                        key,
                        {
                            "file": msg[key][0]["file_id"],
                            "caption": msg.get("caption"),
                            "caption_entities": msg.get("caption_entities"),
                        },
                    )
                elif key in [
                    "animation",
                    "audio",
                    "document",
                    "video",
                    "voice",
                ]:
                    yield MessageSegment(
                        key,
                        {
                            key: msg[key],
                            "caption": msg.get("caption"),
                            "caption_entities": msg.get("caption_entities"),
                        },
                    )
                elif key in ["sticker", "video_note", "dice", "poll"]:
                    yield MessageSegment(key, {key: msg[key]})
            return
        elif isinstance(msg, Iterable) and not isinstance(msg, str):
            for seg in msg:
                yield MessageSegment(seg["type"], seg.get("data") or {})
            return
        elif isinstance(msg, str):
            yield MessageSegment.text(msg)



import json
import asyncio
import pathlib
from typing import Any, List

import aiofiles
from nonebot.log import logger
from nonebot.typing import overrides
from nonebot.message import handle_event
from nonebot.drivers import (
    URL,
    Driver,
    Request,
    Response,
    ForwardDriver,
    ReverseDriver,
    HTTPServerSetup,
)

from nonebot.adapters import Adapter as BaseAdapter

from .bot import Bot
from .event import Event
from .exception import NetworkError
from .config import BotConfig, AdapterConfig


class Adapter(BaseAdapter):
    @overrides(BaseAdapter)
    def __init__(self, driver: Driver, **kwargs: Any):
        super().__init__(driver, **kwargs)
        self.adapter_config = AdapterConfig(**self.config.dict())
        self.tasks: List[asyncio.Task] = []
        self.setup()

    @classmethod
    @overrides(BaseAdapter)
    def get_name(cls) -> str:
        return "Telegram"

    def setup_webhook(self, bot_configs: List[BotConfig]):
        @self.driver.on_startup
        async def _():
            async def handle_http(request: Request) -> Response:
                if request.content:
                    message: dict = json.loads(request.content)
                    await handle_event(bot, Event.parse_event(message))
                return Response(204)

            for bot_config in bot_configs:
                bot = Bot(self, bot_config)
                logger.info("Delete old webhook")
                await self._call_api(bot, "delete_webhook")
                logger.info("Set new webhook")
                await self._call_api(
                    bot,
                    "set_webhook",
                    url=f"{bot.bot_config.webhook_url}/telegram/{bot.self_id}",
                )
                self.bot_connect(bot)
                setup = HTTPServerSetup(
                    URL(f"/telegram/{bot.self_id}"),
                    "POST",
                    self.get_name(),
                    handle_http,
                )
                self.setup_http_server(setup)

    def setup_polling(self, bot_configs: List[BotConfig]):
        @self.driver.on_startup
        async def _():
            async def poll(bot: Bot):
                logger.info("Delete old webhook")
                await self._call_api(bot, "delete_webhook")
                logger.info("Start poll")
                self.bot_connect(bot)

                update_offset = 0
                while True:
                    try:
                        message = (
                            await self._call_api(
                                bot, "get_updates", offset=update_offset
                            )
                        )["result"]
                        if update_offset:
                            for msg in message:
                                if msg["update_id"] > update_offset:
                                    update_offset = msg["update_id"]
                                    await handle_event(bot, Event.parse_event(msg))
                        elif message:
                            update_offset = message[0]["update_id"]
                    except Exception as e:
                        logger.error(e)
                    await asyncio.sleep(bot.bot_config.polling_interval)

            for bot_config in bot_configs:
                self.tasks.append(asyncio.create_task(poll(Bot(self, bot_config))))

        @self.driver.on_shutdown
        async def _():
            for task in self.tasks:
                if not task.done():
                    task.cancel()
            await asyncio.gather(*self.tasks, return_exceptions=True)

    def setup(self) -> None:
        if isinstance(self.driver, ForwardDriver):
            polling_bot_configs = []
            webhook_bot_configs = []

            for bot_config in self.adapter_config.telegram_bots:
                if bot_config.webhook_url:
                    if isinstance(self.driver, ReverseDriver):
                        webhook_bot_configs.append(bot_config)
                    else:
                        raise Exception
                else:
                    polling_bot_configs.append(bot_config)

            self.setup_webhook(webhook_bot_configs)
            self.setup_polling(polling_bot_configs)
        else:
            raise Exception

    @overrides(BaseAdapter)
    async def _call_api(self, bot: Bot, api: str, **data) -> Any:
        # 将方法名称改为驼峰式
        api = api.split("_", maxsplit=1)[0] + "".join(
            s.capitalize() for s in api.split("_")[1:]
        )

        files = {}
        for key, value in data.items():
            if isinstance(value, bytes):
                files[key] = value
            else:
                try:
                    async with aiofiles.open(value, "rb") as f:
                        files[key] = (pathlib.Path(value).name, await f.read())
                except:
                    pass
        for key in files:
            data.pop(key)

        request = Request(
            "POST",
            f"{bot.bot_config.api_server}bot{bot.bot_config.token}/{api}",
            data=data if files else None,
            json=data if not files else None,
            files=files,
            proxy=self.adapter_config.proxy
        )
        if isinstance(self.driver, ForwardDriver):
            try:
                response = await self.driver.request(request)
                if 200 <= response.status_code < 300:
                    if not response.content:
                        raise ValueError("Empty response")
                    result = json.loads(response.content)
                    return result
                raise NetworkError(
                    f"HTTP request received unexpected " f"{response.status_code}"
                )
            except NetworkError:
                raise
            except Exception as e:
                raise NetworkError("HTTP request failed") from e



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot', 'nonebot.adapters.telegram']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.7.0,<0.8.0', 'nonebot2>=2.0.0b1,<3.0.0']

setup_kwargs = {
    'name': 'nonebot-adapter-telegram',
    'version': '0.1.0b0',
    'description': 'Telegram Adapter for NoneBot2',
    'long_description': '<div align="center">\n\t<img width="200" src="docs/logo.png" alt="logo"></br>\n\n# NoneBot Adapter Telegram\n\n_✨ Telegram 协议适配 ✨_\n\n</div>\n\n## Manual\n\n[使用指南](./MANUAL.md)\n\n## TODO\n\n- [ ] Inline Mode Support\n- [x] Multiple Bot Support\n- [ ] All Event (with docstring)\n- [ ] All Method (with docstring)\n- [ ] All MessageSegment (with docstring)',
    'author': 'Jigsaw',
    'author_email': 'j1g5aw@foxmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://v2.nonebot.dev/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.3,<4.0.0',
}


setup(**setup_kwargs)
